def url_pagination(params: dict) -> None:
    if "page" not in params or "per_page" not in params:
        raise ValueError(
            "'page' and 'per_page' parameters are required to requests"
            " with pagination"
        )



"""
Mosqlimate client package.

client library for the Mosqlimate project data platform.
"""
from importlib import metadata as importlib_metadata
from typing import List

from mosqlient.client import PlatformClient as Client  # noqa


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "0.1.0"  # changed by semantic-release


version: str = get_version()
__version__: str = version
__all__: List[str] = []  # noqa: WPS410 (the only __variable__ we use)



from typing import Literal


class Author:
    username: str
    institution: str

    def __init__(self, username: str):
        raise NotImplementedError()


class GithubRepo:
    url: str
    owner: str
    repository: str

    def __init__(self, owner: str, repository: str):
        raise NotImplementedError()


class ModelBase:
    name: str
    description: str
    author: Author
    respository: GithubRepo
    disease: Literal["dengue", "zika", "chikungunya"]
    type: Literal["time", "spatial"]
    time_resolution: Literal["day", "week", "month", "year"]
    ADM_level: Literal[0, 1, 2, 3]

    def __init__(self, id):
        print(id)



import asyncio
from urllib.parse import urlparse
from typing import Literal, Optional, Any

import requests
import nest_asyncio

from mosqlient import Client
from mosqlient.config import API_PROD_URL, API_DEV_URL
from mosqlient.requests import get_all
from mosqlient.errors import ClientError, ModelPostError

nest_asyncio.apply()


def _params(**kwargs) -> dict[str, Any]:
    params = {}
    for k, v in kwargs.items():
        if isinstance(v, (bool, int, float, str)):
            params[k] = str(v)
        elif v is None:
            continue
        else:
            raise TypeError(f"Unknown type f{type(v)}")

    return params


class Model:
    client: Client | None

    def __init__(self, client: Optional[Client] = None):
        self.client = client

    @classmethod
    def get(cls, **kwargs):
        cls._validate_fields(**kwargs)
        params = _params(**kwargs)

        async def fetch_models():
            return await get_all("registry", "models", params)

        if asyncio.get_event_loop().is_running():
            loop = asyncio.get_event_loop()
            future = asyncio.ensure_future(fetch_models())
            return loop.run_until_complete(future)
        return asyncio.run(fetch_models())

    def post(
        self,
        name: str,
        description: str,
        repository: str,
        implementation_language: str,
        disease: Literal["dengue", "chikungunya", "zika"],
        temporal: bool,
        spatial: bool,
        categorical: bool,
        adm_level: Literal[0, 1, 2, 3],
        time_resolution: Literal["day", "week", "month", "year"],
        _env: Literal["dev", "prod"] = "prod"
    ):
        if self.client is None:
            raise ClientError(
                "A Client instance must be provided, please instantiate Model "
                "passing your Mosqlimate's credentials. For more infor about "
                "retrieving or inserting data from Mosqlimate, please see the "
                "API Documentation"
            )
        params = {
            "name": name,
            "description": description,
            "repository": repository,
            "implementation_language": implementation_language,
            "disease": disease,
            "temporal": temporal,
            "spatial": spatial,
            "categorical": categorical,
            "ADM_level": adm_level,
            "time_resolution": time_resolution
        }

        self._validate_fields(**params)
        params = _params(**params)
        base_url = API_DEV_URL if _env == "dev" else API_PROD_URL
        url = base_url + "/".join(("registry", "models")) + "/"
        headers = {"X-UID-Key": self.client.X_UID_KEY}
        resp = requests.post(url, json=params, headers=headers, timeout=60)

        if resp.status_code != 201:
            raise ModelPostError(
                f"POST request returned status code {resp.status_code}"
            )

        return resp

    @staticmethod
    def _validate_fields(**kwargs) -> None:
        ModelFieldValidator(**kwargs)


class ModelFieldValidator:
    FIELDS = {
        "id": (int, str),
        "name": str,
        "description": str,
        "author_name": str,
        "author_username": str,
        "author_institution": str,
        "repository": str,
        "implementation_language": str,
        "disease": str,
        "ADM_level": (str, int),
        "temporal": bool,
        "spatial": bool,
        "categorical": bool,
        "time_resolution": str,
    }
    DISEASES = ["dengue", "zika", "chikungunya"]
    ADM_LEVELS = [0, 1, 2, 3]
    TIME_RESOLUTIONS = ["day", "week", "month", "year"]

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if v is None:
                continue

            if not isinstance(v, self.FIELDS[k]):
                raise TypeError(
                    f"Field '{k}' must have instance of "
                    f"{' or '.join(self.FIELDS[k])}"
                )

            if k == "id":
                if v <= 0:
                    raise ValueError("Incorrect value for field 'id'")

            if k == "description":
                if not v:
                    raise ValueError("A Model description must be provided")
                if len(v) > 500:
                    raise ValueError("Model description too long")

            if k == "repository":
                repo_url = urlparse(v)
                if repo_url.netloc != "github.com":
                    raise ValueError(
                        "'repository' must be a valid GitHub repository"
                    )

            if k == "disease":
                if v == "chik":
                    v = "chikungunya"

                if v not in self.DISEASES:
                    raise ValueError(
                        f"Unkown 'disease'. Options: {self.DISEASES}"
                    )

            if k == "ADM_level":
                v = int(v)
                if v not in self.ADM_LEVELS:
                    raise ValueError(
                        f"Unkown 'ADM_level'. Options: {self.ADM_LEVELS}"
                    )

            if k == "time_resolution":
                if v not in self.TIME_RESOLUTIONS:
                    raise ValueError(
                        "Unkown 'time_resolution'. "
                        f"Options: {self.TIME_RESOLUTIONS}"
                    )



import asyncio
import requests
from urllib.parse import urljoin
from typing import AnyStr, Literal
from itertools import chain

import aiohttp

from mosqlient.config import API_DEV_URL, API_PROD_URL, APPS, APPS_TYPE
from mosqlient.utils import validate


def get(
    app: APPS_TYPE,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    pagination: bool = True,
    timeout: int = 10,
    _env: Literal["dev", "prod"] = "prod"
) -> requests.models.Response:
    if app not in APPS:
        raise ValueError(f"unkown Mosqlimate app. Options: {APPS}")

    if pagination:
        validate.url_pagination(params)

    if not endpoint:
        raise ValueError("endpoint is required")

    base_url = API_DEV_URL if _env == "dev" else API_PROD_URL

    url = urljoin(
        base_url,
        "/".join((str(app), str(endpoint)))
    ) + "/?"

    return requests.get(url, params, timeout=timeout)


async def aget(
    session: aiohttp.ClientSession,
    url: str,
    params: dict[str, str | int | float],
    timeout: int = 10,
    retries: int = 3
) -> dict:
    try:
        if retries < 0:
            raise aiohttp.ClientConnectionError("Too many attempts")
        async with session.get(url, params=params, timeout=timeout) as res:
            if res.status == 200:
                return await res.json()
            if retries == 0:
                raise aiohttp.ClientConnectionError(
                    f"Response status: {res.status}. Reason: {res.reason}"
                )
            await asyncio.sleep(10/(retries + 1))
            await aget(session, url, params, timeout, retries - 1)
    except aiohttp.ServerTimeoutError:
        await asyncio.sleep(8/(retries + 1))
        await aget(session, url, params, timeout, retries - 1)
    raise aiohttp.ClientConnectionError("Invalid request")


async def get_all(
    app: APPS_TYPE,
    endpoint: AnyStr,
    params: dict[str, str | int | float],
    timeout: int = 60,
    _max_per_page: int = 50,
    _env: Literal["dev", "prod"] = "prod"
) -> list[dict]:
    params["page"] = 1
    params["per_page"] = _max_per_page

    base_url = API_DEV_URL if _env == "dev" else API_PROD_URL

    url = urljoin(
        base_url,
        "/".join((str(app), str(endpoint)))
    ) + "/?"

    async with aiohttp.ClientSession() as session:
        first_page = await aget(session, url, params)

    total_pages = first_page['pagination']['total_pages']

    if total_pages == 1:
        return first_page['items']

    async with aiohttp.ClientSession(timeout=timeout) as session:
        tasks = []
        for page in range(2, total_pages + 1):
            params["page"] = page
            task = asyncio.create_task(aget(session, url, params))
            tasks.append(task)
        results = await asyncio.gather(*tasks)
    return list(chain(result['items'] for result in results))



from typing import Iterable, Union, Any


class ValidationErrorBase(Exception):
    """Base exception class for validation errors"""

    def __init__(self, message=None):
        super().__init__(message or "A validation error occurred")


class FieldTypeError(ValidationErrorBase):
    """Exception raised for errors in a specific field."""

    def __init__(self, field: str, _type: Union[Iterable, Any]):
        self.field = field
        super().__init__(
            f"Incorrect type for field: '{field}'. Expected type(s): '{_type}'"
        )

    def __str__(self):
        return f"{self.field}: {self.args[0]}"


class ClientError(Exception):
    def __init__(self, message=None):
        super().__init__(message or "Client error")


class ModelPostError(Exception):
    def __init__(self, message=None):
        super().__init__(message or "registry.Model POST request error")
